#include "../../std.h"
#include <iostream>
#include <fstream>

#ifdef HAVE_SSTREAM
#include <sstream>
#else
#include <strstream>
#endif

#include "../../misc/escaped.h"
/*
#include "../../misc/inttostring.h"
#include "../../misc/stringtoint.h"
#include "../../misc/log_time.h"
#include "../../config/environment.h"
#include "../../auth/auth.h"
*/

#include "../jobs.h"
#include "jobs_soap.nsmap"

class HTTP_JobControl_Handle {
 friend class HTTP_JobControl;
 private:
  plugin_init_t jobplugin_init;
  void* jobplugin_handle;
 public:
  HTTP_JobControl_Handle(const char* path);
  ~HTTP_JobControl_Handle(void);
};

HTTP_JobControl_Handle::HTTP_JobControl_Handle(const char* path):jobplugin_handle(NULL),jobplugin_init(NULL) {
cerr<<"HTTP_JobControl_Handle: path: "<<path<<endl;
  jobplugin_handle=dlopen(path,RTLD_NOW);
  if(!jobplugin_handle) {
    odlog(0)<<dlerror()<<endl;
    odlog(0)<<"Can't load plugin "<<path<<endl;
    return;
  };
  jobplugin_init=(plugin_init_t)dlsym(jobplugin_handle,"init");
  if(jobplugin_init == NULL) {
    olog<<"Plugin "<<path<<" is broken."<<endl;
    dlclose(jobplugin_handle); jobplugin_handle=NULL; return;
  };
}

HTTP_JobControl_Handle::~HTTP_JobControl_Handle(void) {
  if(jobplugin_handle) {
    dlclose(jobplugin_handle); jobplugin_handle=NULL; jobplugin_init=NULL;
  };
}

HTTP_JobControl::HTTP_JobControl(HTTP_Connector *c,const char* uri,HTTP_JobControl_Handle& handle):HTTP_ServiceAdv(c),service_url(uri) {
  /* SOAP */
  soap_init();
  sp.namespaces=jobs_soap_namespaces;
  sp.user=this;
  // Initialize plugin's instance
cerr<<"Calling userspec_t"<<endl;
  userspec.fill(c->identity());
cerr<<"Finished userspec_t"<<endl;
  string jobplugin_config = "end\n";
#ifdef HAVE_SSTREAM
  stringstream cfile(jobplugin_config);
#else
  strstream cfile;
  cfile<<jobplugin_config;
#endif
  if(handle.jobplugin_init) {
    nordugrid_config_loc="";
    unsetenv("NORDUGRID_CONFIG");
    jobplugin=(*(handle.jobplugin_init))(cfile,userspec);
    if(jobplugin == NULL) {
      olog<<"Job control plugin failed to initialise." << endl;
      return;
    };
    /*
    if(jobplugin->acquire() != 1) {
      olog<<"Job plugin's acquire failed (should never happen)." << endl;
      delete jobplugin; jobplugin=NULL;
      return;
    };
    */
  };
}

HTTP_JobControl::~HTTP_JobControl(void) {
  soap_deinit();
  /* if(jobplugin) if(jobplugin->release() == 0) { */
  if(jobplugin) {
    delete jobplugin; jobplugin=NULL;
  };
}

HTTP_Service* jobs_service_creator(HTTP_Connector& c,const char* uri,void* arg) {
  HTTP_JobControl_Handle& handle = *((HTTP_JobControl_Handle*)arg);
  HTTP_JobControl *h = new HTTP_JobControl(&c,uri,handle);
  return h;
}

bool jobs_service_configurator(istream& f,const char* uri,HTTP_Service_Properties &prop) {
  string jobplugin_path;
  for(;!f.eof();) {
    char buf[1024];
    istream_readline(f,buf,sizeof(buf));
    char* p = buf;
    for(;*p;p++) if(!isspace(*p)) break;
    char* command = p;
    if(*command == '#') continue;
    for(;*p;p++) if(isspace(*p)) break;
    int command_len = p-command;
    if(!command_len) continue;
cerr<<"command: "<<command<<endl;
cerr<<"command: "<<command_len<<endl;
    if((command_len==3) && (strncmp(command,"end",command_len)==0)) { break; }
    else if((command_len==9) && (strncmp(command,"jobplugin",command_len)==0)) {
cerr<<"command: jobplugin: "<<p<<endl;
      input_escaped_string(p,jobplugin_path);
cerr<<"command: jobplugin: path: "<<jobplugin_path<<endl;
    };
  };
cerr<<"jobs_service_configurator: jobplugin_path: "<<jobplugin_path<<endl;
  HTTP_JobControl_Handle* handle = new HTTP_JobControl_Handle(jobplugin_path.c_str());
  prop.arg=handle;
  prop.subtree=true;
  return true;
}

static char* soap_strdup_l(soap* sp,const char* s,int l = -1) {
  if(l<0) l=strlen(s);
  char* s_ = (char*)soap_malloc(sp,l+1);
  if(s_ == NULL) return NULL;
  memcpy(s_,s,l); s_[l]=0;
  return s_;
}

#ifdef OLD_GSOAP
#error gSOAP older than 2.5.2 is not supported.
#else
void HTTP_JobControl::soap_methods(void) {
  odlog(3)<<"soap_methods: tag: "<<sp.tag<<endl;
  if(!soap_match_tag(&sp,sp.tag,"nj:createJob")) soap_serve_nj__createJob(&sp);
  else sp.error = SOAP_NO_METHOD;
}
#endif

/*
 * HTTP methods
 */
HTTP_Error HTTP_JobControl::get(const char* uri,int &keep_alive) {
  if(!jobplugin) {
cerr<<"jobplugin is not initialized"<<endl;
    return HTTP_NOT_IMPLEMENTED;
  };
  HTTP_Error err;
  if((err=parse_header(keep_alive)) != HTTP_OK) return err;
cerr<<"Header parsed"<<endl;
  if(failure_parsing || unsupported_option_passed) {
    keep_alive=0; return HTTP_FAILURE;
  };
  // obtain requested name
cerr<<"HTTP_JobControl::get: uri: "<<uri<<endl;
cerr<<"HTTP_JobControl::get: service_url: "<<service_url<<endl;
cerr<<"HTTP_JobControl::get: requested_url: "<<requested_url<<endl;
cerr<<"HTTP_JobControl::get: requested_path: "<<requested_path<<endl;
  const char* name = requested_path.c_str();
  for(;*name;++name) if(*name != '/') break;
  odlog(2)<<"HTTP_JobControl::get: name: "<<name<<endl;
  unsigned int l = strlen(name);
  string jobname = name; 
  int p = jobname.find('/');
  if(p != string::npos) jobname.resize(p);
  bool dir_q = false; if((l > 0) && (name[l-1] == '/')) dir_q=true;
  string path_q(name,dir_q?l-1:l);
  odlog(2)<<"HTTP_JobControl::get: jobname: "<<jobname<<endl;
  odlog(2)<<"HTTP_JobControl::get: dir_q: "<<dir_q<<endl;
  odlog(2)<<"HTTP_JobControl::get: path_q: "<<path_q<<endl;
  string html;
  if(path_q.length() == 0) {
    dir_q=true;
    // Top level - generate table of jobs
    html="<HTML><HEAD>J O B S</HEAD>\n<BODY>\n<TABLE>\n";
    list<DirEntry> dir_list;
    if(jobplugin->readdir(path_q.c_str(),dir_list,DirEntry::full_object_info) != 0) {
      if(c->send_error_response(keep_alive,416,NULL,"Error 416") != 0) {
        odlog(3)<<"failed to send 416"<<endl; keep_alive=0; return HTTP_ERROR;
      };
      return HTTP_OK;
    };
    for(list<DirEntry>::iterator d = dir_list.begin();d!=dir_list.end();++d) {
      if(d->name == "info") continue;
      if(d->name == "new") continue;
      html+="<TR>";
      // Status
      string tmp_s;
      char tmp_buf[256];
      unsigned long long int tmp_l;
      tmp_s="info/"; tmp_s+=d->name; tmp_s+="/status";
      if(jobplugin->open(tmp_s.c_str(),GRIDFTP_OPEN_RETRIEVE,0) != 0) {
        html+="<TD>UNKNOWN<TD/>";
      } else {
        tmp_l=sizeof(tmp_buf)-1;
        if(jobplugin->read((unsigned char*)tmp_buf,0,&tmp_l) != 0) {
          jobplugin->close(true);
          html+="<TD>UNKNOWN<TD/>";
        } else {
          jobplugin->close(true);
          tmp_buf[tmp_l]=0;
          html+="<TD>"; html+=tmp_buf; html+="</TD>";
        };
      };
      // Job name + link
      html+="<TD><A HREF=\""+requested_url+d->name+"/\">"+d->name+"</A></TD>";
      // Link to failure
      tmp_s="info/"; tmp_s+=d->name; tmp_s+="/failure";
      DirEntry tmp_inf;
      if(jobplugin->checkfile(tmp_s,tmp_inf,DirEntry::full_object_info) == 0) {
        html+="<TD><A HREF=\""+requested_url+"info/"+d->name+"/failure\">failed</A></TD>";
      } else {
        html+="<TD></TD>";
      };
      // Link to debug
      html+="<TD><A HREF=\""+requested_url+"info/"+d->name+"/errors\">debug</A></TD>";
      html+="</TR>\n";
    };
    html+="</TABLE>\n</BODY>";
  } else if(dir_q && (jobname == "new")) {
    if(c->send_error_response(keep_alive,404,NULL,"Error 404") != 0) {
      odlog(3)<<"failed to send 404"<<endl; keep_alive=0; return HTTP_ERROR;
    };
    return HTTP_OK;
  } else if(dir_q && (jobname == "info")) {
    if(c->send_error_response(keep_alive,404,NULL,"Error 404") != 0) {
      odlog(3)<<"failed to send 404"<<endl; keep_alive=0; return HTTP_ERROR;
    };
    return HTTP_OK;
  } else if(dir_q) {
    // directory
    list<DirEntry> dir_list;
    if(jobplugin->readdir(path_q.c_str(),dir_list,DirEntry::full_object_info) != 0) {
      if(c->send_error_response(keep_alive,416,NULL,"Error 416") != 0) {
        odlog(3)<<"failed to send 416"<<endl; keep_alive=0; return HTTP_ERROR;
      };
      return HTTP_OK;
    };
    // Form a page
    html="<HTML><HEAD>"; html+=name; html+="</HEAD>\n<BODY>\n";
    for(list<DirEntry>::iterator d = dir_list.begin();d!=dir_list.end();++d) {
      html+="<A HREF=\"";
      html+=requested_url; html+=d->name;
      if(!d->is_file) html+="/";
//  time_t changed;
//  time_t modified;
//  unsigned long long size;
//  uid_t uid;
//  gid_t gid;
//  bool may_rename;  //
//  bool may_delete;  //
//  bool may_create;  // for dirs
//  bool may_chdir;   // for dirs
//  bool may_dirlist; // for dirs
//  bool may_mkdir;   // for dirs
//  bool may_purge;   // for dirs
//  bool may_read;    // for files
//  bool may_append;  // for files
//  bool may_write;   // for files
      html+="\">";
      html+=d->name; html+="</A><BR>\n";
    };
    html+="</BODY>";
  };
  if(dir_q) {
    // So far always return full content
    err=send_header(keep_alive,0,html.length()-1,false,html.length(),
                                              HTTP_Time(-1),HTTP_Time(-1));
    if(err != HTTP_OK) {
      keep_alive=0; return HTTP_ERROR;
    };
    if(c->write(html.c_str(),html.length()) != 0) {
      keep_alive=0; return HTTP_ERROR;
    };
    return HTTP_OK;
  };
  // file
  DirEntry fileinfo;
  if(jobplugin->open(name,GRIDFTP_OPEN_RETRIEVE) != 0) {
    if(c->send_error_response(keep_alive,416,NULL,"Error 416") != 0) {
      odlog(3)<<"failed to send 416"<<endl; keep_alive=0; return HTTP_ERROR;
    };
    return HTTP_OK;
  };
  string name_(name);
  if(jobplugin->checkfile(name_,fileinfo,DirEntry::full_object_info) != 0) {
    jobplugin->close(false);
    if(c->send_error_response(keep_alive,416,NULL,"Error 416") != 0) {
      odlog(3)<<"failed to send 416"<<endl; keep_alive=0; return HTTP_ERROR;
    };
    return HTTP_OK;
  };
  // so far only one range is effective
  uint64_t range_start_ = 0;
  uint64_t range_end_ = (uint64_t)(-1);
  if(range_passed) {
    range_start_=range_start[0];
    range_end_=range_end[0];
    if(range_end_ <= range_start_) {
      odlog(1)<<"range_end <= range_start"<<endl;
      jobplugin->close(false);
      if(c->send_error_response(keep_alive,416,NULL,"Error 416") != 0) {
        odlog(3)<<"failed to send 416"<<endl; keep_alive=0; return HTTP_ERROR;
      };
      return HTTP_OK;
    };
  };
  if(fileinfo.size && (range_start_ >= fileinfo.size)) {
    jobplugin->close(false);
    if(c->send_error_response(keep_alive,416,NULL,"Error 416") != 0) {
      odlog(3)<<"failed to send 416"<<endl; keep_alive=0; return HTTP_ERROR;
    };
    return HTTP_OK;
  };
  // Simple solution: always return <= 1MB of first range
  unsigned char buf[1024*1024];
  unsigned long long int size = sizeof(buf);
  if(range_end_ < (range_start_+size)) size=range_end_-range_start_;
  if(jobplugin->read(buf,range_start_,&size) != 0) {
    jobplugin->close(false);
    if(c->send_error_response(keep_alive,416,NULL,"Error 416") != 0) {
      odlog(3)<<"failed to send 416"<<endl; keep_alive=0; return HTTP_ERROR;
    };
    return HTTP_OK;
  };
  jobplugin->close(false);
  range_end_=range_start_+size;
  err=send_header(keep_alive,range_start_,range_end_,true,fileinfo.size,
                                                 HTTP_Time(-1),HTTP_Time(-1));
  if(err != HTTP_OK) { keep_alive=0; return HTTP_ERROR; };
  if(c->write((const char*)buf,size) != 0) {
    keep_alive=0; return HTTP_ERROR;
  };
  return HTTP_OK;
}

HTTP_Error HTTP_JobControl::put(const char* uri,int &keep_alive) {
  if(!jobplugin) {
cerr<<"jobplugin is not initialized"<<endl;
    return HTTP_NOT_IMPLEMENTED;
  };
  HTTP_Error err;
  if((err=parse_header(keep_alive)) != HTTP_OK) return err;
cerr<<"Header parsed"<<endl;
  if(failure_parsing || unsupported_option_passed) {
    keep_alive=0; return HTTP_FAILURE;
  };
  // obtain requested name
cerr<<"HTTP_JobControl::get: uri: "<<uri<<endl;
cerr<<"HTTP_JobControl::get: service_url: "<<service_url<<endl;
  const char* name = requested_path.c_str();
  for(;*name;++name) if(*name != '/') break;
  odlog(2)<<"HTTP_JobControl::get: name: "<<name<<endl;
  unsigned int l = strlen(name);
  string jobname = name;
  int p = jobname.find('/');
  if(p != string::npos) jobname.resize(p);
  bool dir_q = false; if((l > 0) && (name[l-1] == '/')) dir_q=true;
  string path_q(name,dir_q?l-1:l);
  if(dir_q) {
    // PUT directory?
    keep_alive=0;
    if(c->send_error_response(keep_alive,416,NULL,"Error 416") != 0) {
      odlog(3)<<"failed to send 416"<<endl; return HTTP_ERROR;
    };
    return HTTP_OK;
  };
  if(jobname == "new") {
    keep_alive=0;
    if(c->send_error_response(keep_alive,416,NULL,"Error 416") != 0) {
      odlog(3)<<"failed to send 416"<<endl; return HTTP_ERROR;
    };
    return HTTP_OK;
  };
  uint64_t range_start,range_end;
  if(entity_range_passed) {
    range_start=entity_range_start;
    range_end=entity_range_end+1;
    odlog(2)<<"jobs::put: range_passed: "<<range_start<<"-"<<range_end<<endl;
    if(length_passed && ((range_end-range_start) != length)) {
      keep_alive=0;
      if(send_header(keep_alive,406) != HTTP_OK) {
        odlog(3)<<"failed to send 406"<<endl; return HTTP_ERROR;
      };
      return HTTP_OK;
    };
  } else {
    odlog(2)<<"jobs::put: range not passed"<<endl;
    range_start=0;
    if(length_passed) {
      range_end=length;
    } else {
      odlog(2)<<"jobs::put: length not passed too"<<endl;
      keep_alive=0;
      if(send_header(keep_alive,406) != HTTP_OK) {
        odlog(3)<<"failed to send 406"<<endl; return HTTP_ERROR;
      };
      return HTTP_OK;
    };
  };
  if(jobplugin->open(name,GRIDFTP_OPEN_STORE,range_end) != 0) {
    keep_alive=0;
    if(c->send_error_response(keep_alive,416,NULL,"Error 416") != 0) {
      odlog(3)<<"failed to send 416"<<endl; return HTTP_ERROR;
    };
    return HTTP_OK;
  };
  err=HTTP_OK;
  // receive entity from network and store locally
  for(;;) {
    if(range_start >= range_end) break;
    uint64_t l = range_end-range_start;
    char buf[1024*1024]; // 1MB - is it enough ?
    if(l > sizeof(buf)) l=sizeof(buf);
    // globus_io tends to return no more than SSL block, hence
    // try to fill buffer by looping
    uint64_t ll=0;
    for(;ll<l;) {
      uint64_t l_=c->read(buf+ll,l-ll);
      if(l_ == 0) break; ll+=l_;
    };
    if(ll == 0) break; // EOF
    if(jobplugin->write((unsigned char*)buf,range_start,ll) != 0) {
      odlog(0)<<"Failed to write to file"<<endl;
      err=HTTP_FAILURE;
      break;
    };
  };
  if(err != HTTP_OK) {
    jobplugin->close(false);
    odlog(0)<<"jobs::put: failure writing data from network to disc"<<endl;
    keep_alive=0;
    send_header(keep_alive,500);
    err=HTTP_FAILURE;
  } else {
    jobplugin->close(true);
    if(send_header(keep_alive,200) != HTTP_OK) {
      keep_alive=0;
      odlog(3)<<"failed to send 200"<<endl; 
      err=HTTP_ERROR;
    };
  };
  return err;
}

HTTP_Error HTTP_JobControl::post(const char* uri,int &keep_alive) {
  const char* name = requested_path.c_str();
  for(;*name;++name) if(*name != '/') break;
  int l = strlen(name);
  for(;l;--l) { if(name[l-1] != '/') break; };
  job_id.assign(name,l); job_only=true;
  if(job_id.find('/') != string::npos) job_only=false;
  enum HTTP_Error r = soap_post(uri,keep_alive);
  return r;
}

static int make_failure(struct soap *sp,int code,char* reason,nj__jobinfo& r) {
  r.failure_code=(int*)soap_malloc(sp,sizeof(int));
  if(r.failure_code) *r.failure_code=1;
  r.failure_reason="Not suitable path for this request";
  return SOAP_OK;
}

/*
 *  SOAP interface functions
 */
int nj__createJobs(struct soap *sp, int __size_jobdef, nj__jobdefinition* jobdef, __jobsinfo& r) {
  HTTP_JobControl* it = (HTTP_JobControl*)(sp->user);
  r.__size_job=__size_jobdef;
  r.job=NULL;
  if(r.__size_job == 0) return SOAP_OK;
  r.job=soap_new_nj__jobinfo(sp,r.__size_job);
  if(r.job == NULL) return SOAP_OK;
  if(!it->job_id.empty()) {
    for(int i=0;i<__size_jobdef;i++) {
      make_failure(sp,1,"Not suitable path for this request",r.job[i]);
    };
    return SOAP_OK;
  };
  for(int i=0;i<__size_jobdef;i++) {
    nj__createJob(sp,jobdef+i,r.job[i]);
  };
  return SOAP_OK;
}

int nj__createJob(struct soap *sp,nj__jobdefinition* jobdef,nj__jobinfo& r) {
  HTTP_JobControl* it = (HTTP_JobControl*)(sp->user);
  if(!it->job_id.empty()) {
    return make_failure(sp,1,"Not suitable path for this request",r);
  };
  if(!(it->jobplugin)) {
cerr<<"jobplugin is not initialized"<<endl;
    return make_failure(sp,1,"Internal configuration error",r);
  };
  if(jobdef == NULL) {
    return make_failure(sp,1,"Missing job",r);
  };
  if(jobdef->description == NULL) {
    return make_failure(sp,1,"Missing job",r);
  };
  int l = strlen(jobdef->description);
  if(l == 0) {
    return make_failure(sp,1,"Missing job",r);
  };
  // simulate RSL submission to jobplugin
  it->job_id="new";
  if(it->jobplugin->checkdir(it->job_id) != 0) {
    return make_failure(sp,1,"Internal error",r);
  };
  if(it->jobplugin->open("new/job",GRIDFTP_OPEN_STORE,l) != 0) {
    return make_failure(sp,1,"Internal error",r);
  };
  if(it->jobplugin->write((unsigned char*)(jobdef->description),0,l) != 0) {
    it->jobplugin->close(false);
    return make_failure(sp,1,"Internal error",r);
  };
  if(it->jobplugin->close(true) != 0) {
    return make_failure(sp,1,"RSL not accepted",r);
  };
  make_failure(sp,0,NULL,r);
  r.state="ACCEPTED";
  r.id=(char*)(it->job_id.c_str()); // This variable is valid till next SOAP message
  string u(it->requested_url.c_str()); u+="/"; u+=it->job_id; 
  r.url=soap_strdup_l(sp,u.c_str());
  r.def=NULL;
  r.__size_attributes=0;
  r.attributes=NULL;
  return SOAP_OK;
}

int nj__queryJob(struct soap *sp, nj__jobinfo& r) {
  HTTP_JobControl* it = (HTTP_JobControl*)(sp->user);
  if(it->job_id.empty() || !it->job_only) {
    r.failure_code=(int*)soap_malloc(sp,sizeof(int));
    if(r.failure_code) *r.failure_code=1;
    r.failure_reason="Not suitable path for this request";
    return SOAP_OK;
  };
  r.failure_code=(int*)soap_malloc(sp,sizeof(int));
  if(r.failure_code) *r.failure_code=1;
  r.failure_reason="Not implemented";
  return SOAP_OK;
}

int nj__cancelJob(struct soap *sp, nj__jobinfo& r) {
  HTTP_JobControl* it = (HTTP_JobControl*)(sp->user);
  if(it->job_id.empty() || !it->job_only) {
    r.failure_code=(int*)soap_malloc(sp,sizeof(int));
    if(r.failure_code) *r.failure_code=1;
    r.failure_reason="Not suitable path for this request";
    return SOAP_OK;
  };
  r.failure_code=(int*)soap_malloc(sp,sizeof(int));
  if(r.failure_code) *r.failure_code=1;
  r.failure_reason="Not implemented";
  return SOAP_OK;
}

int nj__cleanJob(struct soap *sp, nj__jobinfo& r) {
  HTTP_JobControl* it = (HTTP_JobControl*)(sp->user);
  if(it->job_id.empty() || !it->job_only) {
    r.failure_code=(int*)soap_malloc(sp,sizeof(int));
    if(r.failure_code) *r.failure_code=1;
    r.failure_reason="Not suitable path for this request";
    return SOAP_OK;
  };
  r.failure_code=(int*)soap_malloc(sp,sizeof(int));
  if(r.failure_code) *r.failure_code=1;
  r.failure_reason="Not implemented";
  return SOAP_OK;
}

int nj_modifyJob(struct soap *sp, nj__jobinfo* job, nj__jobinfo& r) {
  HTTP_JobControl* it = (HTTP_JobControl*)(sp->user);
  if(it->job_id.empty() || !it->job_only) {
    r.failure_code=(int*)soap_malloc(sp,sizeof(int));
    if(r.failure_code) *r.failure_code=1;
    r.failure_reason="Not suitable path for this request";
    return SOAP_OK;
  };
  r.failure_code=(int*)soap_malloc(sp,sizeof(int));
  if(r.failure_code) *r.failure_code=1;
  r.failure_reason="Not implemented";
  return SOAP_OK;
}

